/*

kHTTPd -- the next generation

Send actual file-data to the connections

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

Purpose:

DataSending does the actual sending of file-data to the socket.

Note: Since asynchronous reads do not -yet- exists, this might block!

Return value:
	The number of requests that changed status
*/

#include <linux/kernel.h>
#include <linux/skbuff.h>

#include <net/tcp.h>

#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#include "structure.h"
#include "prototypes.h"
#include "config.h"

STATIC	char	*Block[KHTTPD_NUMCPU];


int DataSending(int CPU)
{
	struct http_request *CurrentRequest,**Prev;
	int count = 0;
	
	EnterFunction("DataSending");
	
	Prev = &(threadinfo[CPU].DataSendingQueue);
	CurrentRequest = threadinfo[CPU].DataSendingQueue;
	while (CurrentRequest!=NULL)
	{
#ifndef SENDFILE	
		int ReadSize;
		int retval;
		mm_segment_t oldfs;


		/* This is NOT fast. In fact, it is lame. */
		
		ReadSize = min(4096,CurrentRequest->FileLength - CurrentRequest->BytesSent);
		
		if (ReadSize>0)
		{
		
			CurrentRequest->filp->f_pos = CurrentRequest->BytesSent;
		
			oldfs = get_fs(); set_fs(KERNEL_DS);
			retval = CurrentRequest->filp->f_op->read(CurrentRequest->filp, Block[CPU], ReadSize, &CurrentRequest->filp->f_pos);
			set_fs(oldfs);
		
			if (retval>0)
			{
				retval = SendBuffer_async(CurrentRequest->sock,Block[CPU],(size_t)retval);
				if (retval>0)
				{
					CurrentRequest->BytesSent += retval;
					count++;
				
				}
			}
		
		}
		lock_sock(CurrentRequest->sock->sk);
		if ((CurrentRequest->BytesSent>=CurrentRequest->FileLength)||
		   (CurrentRequest->sock->sk->state !=TCP_ESTABLISHED))
#else
		if ((CurrentRequest->Done!=0)||
		   ((CurrentRequest->sock->sk!=NULL)&&(CurrentRequest->sock->sk->state !=TCP_ESTABLISHED)))
#endif		   
		{
			struct http_request *Next;
			Next = CurrentRequest->Next;
			
			release_sock(CurrentRequest->sock->sk);
			
			if (CurrentRequest->sock->sk->state ==TCP_ESTABLISHED)
			{
				CurrentRequest->sock->sk->nonagle = 0;
				CurrentRequest->sock->sk->linger = 1;
				tcp_push_pending_frames(CurrentRequest->sock->sk,&(CurrentRequest->sock->sk->tp_pinfo.af_tcp));
			}	


			(*Prev) = CurrentRequest->Next;
			
			CurrentRequest->Next = threadinfo[CPU].LoggingQueue;
			threadinfo[CPU].LoggingQueue = CurrentRequest;	
				
			CurrentRequest = Next;
			continue;
		
		}
		release_sock(CurrentRequest->sock->sk);
		

		Prev = &(CurrentRequest->Next);	
		CurrentRequest = CurrentRequest->Next;
	}
	
	LeaveFunction("DataSending");
	return count;
}

int InitDataSending()
{
	int I,I2;
	
	EnterFunction("InitDataSending");
	I=0;
	while (I<KHTTPD_NUMCPU)
	{
		Block[I] = (char*)get_free_page((int)GFP_KERNEL);
		if (Block[I] == NULL) 
		{
			I2=0;
			while (I2<KHTTPD_NUMCPU-1)
			{
				free_page((unsigned long)Block[I2++]);
			}
			LeaveFunction("InitDataSending - abort");
			return -1;
		};
		I++;
	}
	LeaveFunction("InitDataSending");
	return 0;		
}

void StopDataSending(int CPU)
{
	struct http_request *CurrentRequest;
	
	EnterFunction("StopDataSending");
	CurrentRequest = threadinfo[CPU].DataSendingQueue;

	while (CurrentRequest!=NULL)
	{	
		CleanUpRequest(CurrentRequest);
		CurrentRequest=CurrentRequest->Next;		
	}

	free_page( (unsigned long)Block[CPU]);
	LeaveFunction("StopDataSending");
}