#ifdef SENDFILE
/*

kHTTPd -- the next generation

Asynchronous sendfile daemon

*/
/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/

/*

*/

#include <linux/kernel.h>
#include <linux/skbuff.h>
#include <net/tcp.h>
#include <asm/uaccess.h>
#include <linux/malloc.h>
#include <linux/shm.h>
#include <linux/mman.h>
#include <linux/locks.h>
#include <linux/pagemap.h>
#include <linux/swap.h>
#include <linux/smp_lock.h>
#include <linux/blkdev.h>
#include <linux/file.h>
#include <linux/swapctl.h>
#include <linux/slab.h>
#include <linux/mm.h>

#include <asm/pgtable.h>

#include "structure.h"
#include "prototypes.h"


STATIC struct async_sendfile *SendfileQueue[KHTTPD_NUMCPU]; 
STATIC struct semaphore async_sendfile_mutex[KHTTPD_NUMCPU];

/*@null@*/STATIC struct wait_queue  *DummyWQ=NULL;

STATIC struct  task_struct *SendfileD[KHTTPD_NUMCPU];

STATIC int TryFile(struct async_sendfile *File);

#ifdef DEBUG
#define DebugPrintk(x...)  printk(x)
#else
#define DebugPrintk(x) {}
#endif

STATIC int SendfileCounter=0;

void InitSendfile()
{
#ifndef __LCLINT__
	int I;
	
	EnterFunction("InitSendfile");
	
	I=0;
	while (I<KHTTPD_NUMCPU)
	{
		async_sendfile_mutex[I]=MUTEX;
		SendfileQueue[I]=NULL;
		I++;
	}
	
#endif	
	LeaveFunction("InitSendfile");
}

int Sendfile_Daemon(void* unused)
{
	struct async_sendfile *Current,**Prev,*Temp;
	int changes = 0;
	int count=0;
	int CPU=0;
	sigset_t tmpsig;
	
	if (unused!=NULL)
	  CPU=*( (int*)unused);
	
	exit_mm(current);
	sprintf(current->comm,"kasyncd %i",CPU);

	/* Block all signals except SIGKILL, SIGSTOP and SIGHUP */
	spin_lock_irq(&current->sigmask_lock);
	tmpsig = current->blocked;
	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP)| sigmask(SIGHUP));
	recalc_sigpending(current);
	spin_unlock_irq(&current->sigmask_lock);
	
	SendfileD[CPU]=current;
	
	

	
	while (sysctl_khttpd_stop == 0)
	{
		down(&async_sendfile_mutex[CPU]);
		
		Current = SendfileQueue[CPU];
		Prev = &SendfileQueue[CPU];
		
		changes = 0;
		while (Current!=NULL)
		{
/*			printk("Pos = %i,   Length = %i \n",*(Current->Pos),*(Current->Length));*/

			if ((Current->Pos >= Current->Length)||(Current->PermanentError!=0))
			{ /* Request is done, dequeue it */
		
				DebugPrintk("Request done\n");
				*Prev = Current->Next;
				Temp = Current->Next;
				
				*(Current->Done) = 1;
				
				if (Current->page_cache!=0)
				 	page_cache_free(Current->page_cache);
				 	
				if ((Current->OnWaitQueue!=0)&&(Current->page!=NULL))
				{
					remove_wait_queue(&(Current->page->wait),&(Current->sleep));	
					Current->OnWaitQueue=0;
				}
				 	
				 DebugPrintk("WakeUp\n");	
/*				if (!Current->clientsleep)
				  wake_up_process(Current->clientsleep->task);*/
				if (Current->sock->sk!=NULL)
					wake_up(Current->sock->sk->sleep);
				kfree(Current);
				Current = Temp;
				continue;
			}
			DebugPrintk("End of loop\n");
			
			
			changes += TryFile(Current);
			
			Prev = &(Current->Next);
			Current = Current->Next;
		}
		up(&async_sendfile_mutex[CPU]);
		
		count++;
		current->state = TASK_INTERRUPTIBLE;	
		run_task_queue(&tq_disk);
		if ((changes==0)||(count>100)) 
		{
			count=0;
			(void)interruptible_sleep_on_timeout(&DummyWQ,1);	
		}
		
		if (signal_pending(current)!=0)
			break;

	}
	
	
	(void)printk("Ended asynchronous Sendfile daemon number %i\n",CPU);
	return 0;
}

int AddSendFile(struct async_sendfile *Data,/*@unused@*/ struct wait_queue *WQ)
{
	int CPU=0;
	
	EnterFunction("AddSendFile");
	
	CPU = (SendfileCounter++) % KHTTPD_NUMCPU;
		
	down(&async_sendfile_mutex[CPU]);
	
	Data->Next = SendfileQueue[CPU];
	SendfileQueue[CPU] = Data;
	
	Data->inode = Data->filp->f_dentry->d_inode;
	
	Data->sleep.task = SendfileD[CPU];
/*	Data->clientsleep = WQ;*/
	
	up(&async_sendfile_mutex[CPU]);
	
	wake_up_process(SendfileD[CPU]); /* This doesn't work. */
	
	LeaveFunction("AddSendFile");
	return 0; /* Should return "ticket number" later */
}

STATIC inline int get_max_readahead(struct inode * inode)
{
	if ((inode->i_dev!=0) || (max_readahead[MAJOR(inode->i_dev)]!=0))
		return MAX_READAHEAD;
	return max_readahead[MAJOR(inode->i_dev)][MINOR(inode->i_dev)];
}

STATIC inline void add_to_page_cache(struct page * page,
	struct inode * inode, unsigned long offset,
	struct page **hash)
{
	atomic_inc(&page->count);
	page->flags = (page->flags & ~((1 << PG_uptodate) | (1 << PG_error))) | (1 << PG_referenced);
	page->offset = offset;
	add_page_to_inode_queue(inode, page);
	__add_page_to_hash_queue(page, hash);
}


STATIC int TryFile(struct async_sendfile *File)
{
	size_t	pos;
	int	error,offset,size,size2;
	int	WorkDone = 0;

	EnterFunction("TryFile");

	pos = (File->Pos);
	
	/* Step 1: If we haven't found the page, try to find it now */
	DebugPrintk("TryFile: Step 1 \n");
	
	if (!File->page)	
	{
		File->hash = page_hash(File->inode,pos & PAGE_CACHE_MASK);
		File->page = __find_page(File->inode,pos & PAGE_CACHE_MASK, *(File->hash));
	}
	
	
	if (!File->page)  /* We can't find the page, so we have to allocate it */
	{
		/* Step 2: Allocate a spare page */
		DebugPrintk("TryFile: Step 2 \n");
		if (File->page_cache!=0)
		{
			File->page_cache = page_cache_alloc();
			/* This may have slept. We should start all over; we have
			   made some progress so we return positive */
			return (File->page_cache);
		}
		
		/* Step 3: Add the page to the page-cache */
		DebugPrintk("TryFile: Step 3 \n");
		
		File->page = page_cache_entry(File->page_cache);
		File->page_cache = 0;
		
		add_to_page_cache(File->page,File->inode,pos & PAGE_CACHE_MASK,File->hash);

	
		/* Step 4: Issue the read-request on the page */
		DebugPrintk("TryFile: Step 4 \n");
		
		File->filp->f_ramax = (unsigned long int)get_max_readahead(File->inode);		   
		error = File->inode->i_op->readpage(File->filp,File->page);
		if (error!=0) goto HaveError;
		WorkDone++;
		}
	/* Step 5: Wait for the page to become available and valid */
	DebugPrintk("TryFile: Step 5 \n");
	
	if (PageLocked(File->page)!=0)
	{
		if ((!File->page->wait)&&(File->OnWaitQueue==0))
		{
		  File->OnWaitQueue = 1;
		  add_wait_queue(&(File->page->wait),&(File->sleep));
		}
		return WorkDone;   
	}
	
	if (!PageUptodate(File->page)!=0) /* Woops.... re-read it */
	{
		(void)printk("TryFile: Step 5 - reread \n");
		error = File->inode->i_op->readpage(File->filp,File->page);
		if (error!=0) goto HaveError;
		return WorkDone;	
	}
	
	/* Step 6: Send the data */
	DebugPrintk("TryFile: Step 6 \n");
	
	offset = pos & ~PAGE_CACHE_MASK;
	size   = PAGE_CACHE_SIZE - offset;
	size = min(size, (File->inode->i_size - pos)); /* FIXME!! */
	
	size2 = SendBuffer_async(File->sock,(const char*) (page_address(File->page)+offset),size);
	if ((size2>0)||(size2!=-EAGAIN))
	{
		(File->Pos) += size2;
		WorkDone++;
		
		if (size2<0)
			File->PermanentError = 1;
		if ((size2 == size)||(size2<0)) /* All of the page is sent, or error */
		{
		
			if ((File->OnWaitQueue)!=0)
			{
				remove_wait_queue(&(File->page->wait),&(File->sleep));
				File->OnWaitQueue=0;
			}
			page_cache_release(File->page);  
			File->page = NULL;
		}
	}
	    	
	/* Step 7: Perform readahead if needed */
	DebugPrintk("TryFile: Step 7 \n");
	
	
	/* ReadAhead not implemented for now */


	LeaveFunction("TryFile - success");	
	return WorkDone;
	
	
HaveError:
	if ((File->OnWaitQueue!=0)&&(File->page!=NULL))
	{
		remove_wait_queue(&(File->page->wait),&(File->sleep));
		File->OnWaitQueue=0;
	}
	if (File->page!=NULL)
		page_cache_release(File->page);
	
	return 0;
}



#endif
